# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Jinja template loading."""

import shutil
import tempfile
from os import environ
from os.path import dirname, join

import pytest
from flask import Blueprint, render_template
from jinja2 import TemplateNotFound

from invenio_app.factory import create_ui


#
# Fixtures
#
@pytest.fixture()
def instance_path():
    """Fixture for creating an instance path."""
    path = tempfile.mkdtemp()
    environ.update(
        INVENIO_INSTANCE_PATH=environ.get("INSTANCE_PATH", path),
    )
    yield path
    environ.pop("INVENIO_INSTANCE_PATH", None)
    shutil.rmtree(path)


@pytest.fixture()
def blueprint():
    """Blueprint for loading templates."""
    # This blueprint picks up the tests/templates/ folder, so that we can
    # render templates from inside this folder.
    return Blueprint("tests", __name__, template_folder="templates")


@pytest.fixture()
def notheme_app(instance_path, blueprint):
    """Application without template theming."""
    app = create_ui()
    app.register_blueprint(blueprint)
    with app.app_context():
        yield app


@pytest.fixture()
def theme_app(instance_path, blueprint):
    """Application with template theming."""
    app = create_ui(APP_THEME=["semantic-ui", "bootstrap3"])
    app.register_blueprint(blueprint)
    with app.app_context():
        yield app


@pytest.fixture()
def theme_app_instance_templates(theme_app):
    """Copy templates into instance folder."""
    # Fixture depends on theme_app to ensure the correct instance path is used.
    src = join(dirname(__file__), "instance/templates")
    dst = join(theme_app.instance_path, "templates")
    shutil.copytree(src, dst)
    yield src
    shutil.rmtree(dst)


#
# Tests
#
def test_notheme(notheme_app):
    """Test template loading order *without* themes."""
    assert render_template("base.html") == "base"
    assert render_template("fallback.html") == "fallback-base"
    assert render_template("only.html") == "only-base"

    # Theme templates can also be accessed via full path (not true for base)
    assert render_template("semantic-ui/base.html") == "semantic-ui"
    assert render_template("bootstrap3/base.html") == "bootstrap3"
    assert render_template("bootstrap3/fallback.html") == "fallback-bootstrap3"

    # Not found template
    assert pytest.raises(TemplateNotFound, render_template, "invalid.html")


def test_theme(theme_app):
    """Test template loading order *with* themes configured."""
    # Primary theme defines base.html
    assert render_template("base.html") == "semantic-ui"
    # Primary theme does not define fallback.html so using fallback
    assert render_template("fallback.html") == "fallback-bootstrap3"
    # Primary and secondary theme does not define only.html so using fallback
    # to normal loading
    assert render_template("only.html") == "only-base"

    # Theme templates can also be accessed via full path (not true for base)
    assert render_template("semantic-ui/base.html") == "semantic-ui"
    assert render_template("bootstrap3/base.html") == "bootstrap3"

    # Not found template
    assert pytest.raises(TemplateNotFound, render_template, "invalid.html")


def test_theme_with_instance_templates(theme_app_instance_templates):
    """Test template loading order *with* themes configured."""
    assert render_template("instance.html") == "instance-only-semantic-ui"
    # Primary theme defines base.html
    assert render_template("base.html") == "instance-semantic-ui"
    # Primary theme does not define fallback.html so using fallback
    assert render_template("fallback.html") == "instance-fallback-bootstrap3"
    # Primary and secondary theme does not define only.html so using fallback
    # to normal loading
    assert render_template("only.html") == "instance-only-base"

    # Theme templates can also be accessed via full path (not true for base)
    assert render_template("semantic-ui/base.html") == "instance-semantic-ui"
    assert render_template("bootstrap3/base.html") == "instance-bootstrap3"

    # Instance does not defined noinstance.html
    assert render_template("noinstance.html") == "noinstance-semantic-ui"

    # Not found template
    assert pytest.raises(TemplateNotFound, render_template, "invalid.html")


def test_list_templaes(theme_app):
    """Test list templates."""
    assert sorted(theme_app.jinja_env.loader.list_templates()) == [
        "base.html",
        "bootstrap3/base.html",
        "bootstrap3/fallback.html",
        "fallback.html",
        "only.html",
        "semantic-ui/base.html",
        "semantic-ui/noinstance.html",
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App configuration.

`Invenio-App` is partially overwriting default configuration of `Limiter` and
`Talisman` applications.
You can find below more details about which configuration are set.

For more information, please also see
`Flask-Limiter <https://flask-limiter.readthedocs.io/en/stable/>`_ and
`Flask-Talisman <https://github.com/GoogleCloudPlatform/flask-talisman/>`__
websites.
"""

from invenio_app.limiter import set_rate_limit

RATELIMIT_APPLICATION = set_rate_limit
"""Global rate limit."""

RATELIMIT_STRATEGY = "moving-window"
"""The rate limiting strategy to use.

The strategy used here is the most consistant but also expensive one.
If you are experiencing performance issues due to the increased Redis
traffic, you can replace it with another one from the following
`Flask-Limiter strategies
<https://flask-limiter.readthedocs.io/en/stable/#ratelimit-strategy>`_.
"""

RATELIMIT_HEADERS_ENABLED = True
"""Enable rate limit headers. (Default: ``True``)"""

RATELIMIT_STORAGE_URI = "memory://"
"""Storage backend to store rate-limiting information.

    Memory is used by default if no value is provided.
    For more information regarding the mentioned above configuration values and
    their available options you can see the `Flask-Limiter configuration
    <https://flask-limiter.readthedocs.io/en/stable/#configuration>`_.

.. note::

   Provide your Redis URL if you are rate limiting a multithreaded application.

"""

RATELIMIT_KEY_FUNC = None
"""Define custom key function.

This config is not part of Flask-Limiter.

This function is used to generate a unique key for each visitor to track
the number of performed requests. If not defined, the default ``key_func``
will be used, which will create the key by concatenating the user agent and
the IP address of the user.

For more information you can also see `here
<https://flask-limiter.readthedocs.io/en/stable/#rate-limit-domain>`_
"""

RATELIMIT_PER_ENDPOINT = {}
"""Specifically defined Flask rate limits per endpoint.

This config is not part of Flask-Limiter.
Use this for endpoints that need to be *whitelisted*, providing the Flask
blueprint path accompanied by a `rate limit value
<https://flask-limiter.readthedocs.io/en/stable/#rate-limit-string-notation>`_.

.. code-block:: python

    RATELIMIT_PER_ENDPOINT = \
    {
        'zenodo_frontpage.index': '10 per second',
        'security.login': '10 per second'
    }
"""

RATELIMIT_AUTHENTICATED_USER = "5000 per hour;100 per minute"
"""Rate limit for logged in users."""

RATELIMIT_GUEST_USER = "1000 per hour;60 per minute"
"""Rate limit for non logged in users."""

APP_THEME = None
"""Application-wide themes list used for template and assets lookup.

The value is a list of theme strings applied in a fallback fashion in the order
they are specified:

.. code-block:: python

    APP_THEME = ['my-overlay', 'semantic-ui']

From the above example, templates and assets with the ``my-overlay`` prefix
will be looked up first, and if not found the ``semantic-ui`` prefix will be
used. If none of the lookups are successful, a non-prefixed lookup is done.
"""

APP_ENABLE_SECURE_HEADERS = True
"""Enable Secure Headers. (Default: ``True``)

In case you want to disable completely `Talisman`, you can set to `False`.

Remember that, for development purpose, setting ```DEBUG = True``` is already
enough to disable any side effects such as force ``https``.

.. note::
    `W3C
    <https://www.w3.org/TR/CSP2/>`_
"""

APP_DEFAULT_SECURE_HEADERS = {
    "force_https": True,
    "force_https_permanent": False,
    "force_file_save": False,
    "frame_options": "sameorigin",
    "frame_options_allow_from": None,
    "strict_transport_security": True,
    "strict_transport_security_preload": False,
    "strict_transport_security_max_age": 31556926,  # One year in seconds
    "strict_transport_security_include_subdomains": True,
    "content_security_policy": {"default-src": ["'self'"], "object-src": ["'none'"]},
    "content_security_policy_report_uri": None,
    "content_security_policy_report_only": False,
    "session_cookie_secure": True,
    "session_cookie_http_only": True,
}
"""Talisman default Secure Headers configuration.

As default, invenio assumes that HTTPS is enabled.
If you are not using SSL, then remember to disable the `force_https` and
`session_cookie_secure` configuration options related to HTTPS.

Please note that, as default `Talisman` behaviour, if Flask `DEBUG` mode is on,
then also many security barriers are automatically switched off
(e.g. `force_https` and `session_cookie_secure`).

.. note:: Overwrite
    `Flask-Talisman
    <https://github.com/GoogleCloudPlatform/flask-talisman>`_ configuration.

.. code-block:: python

    from flask import Flask
    from flask_talisman import Talisman

    app = Flask(__name__)
    app.config.update(
        SECRET_KEY='SECRET_KEY'
    )
    talisman = Talisman(app)

    @app.route('/defenders')
    @talisman(frame_options_allow_from='*')
    def defenders():
        \"\"\"Override policies for the specific view.\"\"\"
        return 'Jessica Jones'
"""

APP_ALLOWED_HOSTS = None
"""A list of host/domain names that can be served.

This is a security measure to prevent HTTP Host header attacks, which are
possible even under many seemingly-safe web server configurations.

By default all hosts are allowed. Values in this list can be fully qualified
names (e.g. 'www.example.com'). The validation only applies to
``request.host``.

In addition to this configuration variable, you should make sure that your
web server does not route requests to the application with an invalid Host
header.
"""

APP_HEALTH_BLUEPRINT_ENABLED = True
"""Enable the ping (healthcheck) blueprint. (Default: ``False``)
"""

APP_REQUESTID_HEADER = "X-Request-Id"
"""Name of header containing a request id (max length 200 characters).

If set, the request id will be extracted from the header and set on the global
Flask ``g`` request object. The extracted request id can be used by other
Invenio modules - e.g. Invenio-Logging could include it in log messages.

The request id can be used to trace requests between systems to make
troubleshooting easier.

You can configure Nginx 1.10+ to automatically generate a request id and
add it as a header to both the upstream WSGI server and downstream client::

    add_header X-Request-ID $request_id;

Set to ``None`` to not extract a request id.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CLI test."""

from click.testing import CliRunner


def test_basic_cli():
    """Test version import."""
    from invenio_app.cli import cli

    res = CliRunner().invoke(cli)
    assert res.exit_code == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""CLI application for Invenio flavours."""

from invenio_base.app import create_cli

from .factory import create_app

#: Invenio CLI application.
cli = create_cli(create_app=create_app)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

from invenio_app import InvenioApp


def test_rate_secure_headers(app):
    """Test Rate Limiter extension."""
    app.config["APP_ENABLE_SECURE_HEADERS"] = False
    # Initialize the app
    InvenioApp(app)
    assert "talisman" not in app.extensions


def test_headers(app_with_no_limiter):
    """Test headers."""
    app_with_no_limiter.config["RATELIMIT_APPLICATION"] = "1/day"
    ext = InvenioApp(app_with_no_limiter)

    for handler in app_with_no_limiter.logger.handlers:
        ext.limiter.logger.addHandler(handler)

    @app_with_no_limiter.route("/jessica_jones")
    def jessica_jones():
        return "jessica jones"

    @app_with_no_limiter.route("/avengers")
    def avengers():
        return "infinity war"

    with app_with_no_limiter.test_client() as client:
        res = client.get("/jessica_jones")
        assert res.status_code == 200
        assert res.headers["X-RateLimit-Limit"] == "1"
        assert res.headers["X-RateLimit-Remaining"] == "0"
        assert res.headers["X-RateLimit-Reset"]

        res = client.get("/jessica_jones")
        assert res.status_code == 429
        assert res.headers["X-RateLimit-Limit"]
        assert res.headers["X-RateLimit-Remaining"]
        assert res.headers["X-RateLimit-Reset"]

        res = client.get("/avengers")
        assert res.status_code == 429
        assert res.headers["X-Content-Type-Options"]
        assert res.headers["X-Frame-Options"]
        assert res.headers["X-XSS-Protection"]
        assert res.headers["X-RateLimit-Limit"]
        assert res.headers["X-RateLimit-Remaining"]
        assert res.headers["X-RateLimit-Reset"]


def _normalize_csp_header(header):
    """Normalize a CSP header for consistent comparisons."""
    return {p.strip() for p in (header or "").split(";")}


def _test_csp_default_src(app, expect):
    """Assert that the Content-Security-Policy header is the expect param."""
    ext = InvenioApp(app)

    @app.route("/captain_america")
    def captain_america():
        return "captain america"

    with app.test_client() as client:
        res = client.get("/captain_america")
        assert res.status_code == 200
        assert _normalize_csp_header(
            res.headers.get("Content-Security-Policy")
        ) == _normalize_csp_header(expect)


def test_csp_default_src_when_debug_false(app):
    """Test the Content-Security-Policy header when app debug is False."""
    app.config["DEBUG"] = False
    expect = "default-src 'self'; object-src 'none'"
    _test_csp_default_src(app, expect)


def test_csp_default_src_when_debug_true(app):
    """Test the Content-Security-Policy header when app debug is True."""
    app.config["DEBUG"] = True
    expect = "default-src 'self' 'unsafe-inline'; object-src 'none'"
    _test_csp_default_src(app, expect)


def test_empty_csp_when_set_empty(app):
    """Test empty Content-Security-Policy header when set emtpy."""
    app.config["DEBUG"] = True
    app.config["APP_DEFAULT_SECURE_HEADERS"]["content_security_policy"] = {}
    expect = None
    _test_csp_default_src(app, expect)


def test_default_health_blueprint(app):
    app.config["APP_HEALTH_BLUEPRINT_ENABLED"] = True
    # Initialize the app
    InvenioApp(app)
    with app.test_client() as client:
        res = client.get("/ping")
        assert res.status_code == 200


def test_ping_exempt_from_rate_limiting(app_with_no_limiter):
    app_with_no_limiter.config["APP_HEALTH_BLUEPRINT_ENABLED"] = True
    app_with_no_limiter.config["RATELIMIT_APPLICATION"] = "1/day"
    # Initialize the app
    InvenioApp(app_with_no_limiter)
    with app_with_no_limiter.test_client() as client:
        res = client.get("/ping")
        assert res.status_code == 200
        res = client.get("/ping")
        assert res.status_code == 200
        res = client.head("/ping")
        assert res.status_code == 200
        res = client.options("/ping")
        assert res.status_code == 200


def test_requestid(base_app):
    """Test extraction of header id."""
    InvenioApp(base_app)
    with base_app.test_client() as client:
        assert "1234" == client.get(
            "/requestid", headers={"X-Request-ID": "1234"}
        ).get_data(as_text=True)


def test_requestid_different_header(base_app):
    """Test changing header name."""
    base_app.config["APP_REQUESTID_HEADER"] = "Request-ID"
    InvenioApp(base_app)
    with base_app.test_client() as client:
        # Extracted
        assert "1234" == client.get(
            "/requestid", headers={"Request-ID": "1234"}
        ).get_data(as_text=True)
        # Not extracted
        assert "" == client.get(
            "/requestid", headers={"X-Request-ID": "1234"}
        ).get_data(as_text=True)


def test_requestid_cap_200(base_app):
    """Test cap at 200 chars of request id."""
    InvenioApp(base_app)
    with base_app.test_client() as client:
        assert "1" * 200 == client.get(
            "/requestid", headers={"X-Request-ID": "1" * 500}
        ).get_data(as_text=True)


def test_requestid_no_extraction(base_app):
    """Test no extraction of header id."""
    base_app.config["APP_REQUESTID_HEADER"] = None
    InvenioApp(base_app)
    with base_app.test_client() as client:
        assert "" == client.get(
            "/requestid", headers={"X-Request-ID": "1234"}
        ).get_data(as_text=True)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2019 CERN.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio app extension."""

import warnings

from flask import Blueprint, g, request
from flask_limiter import Limiter
from flask_talisman import Talisman

from invenio_app.limiter import useragent_and_ip_limit_key

from . import config
from .helpers import ThemeJinjaLoader, obj_or_import_string, safe_redirect


class InvenioApp(object):
    """Invenio app extensions."""

    def __init__(self, app=None, **kwargs):
        r"""Extension initialization.

        :param app: An instance of :class:`~flask.Flask`.
        :param \**kwargs: Keyword arguments are passed to ``init_app`` method.
        """
        self.limiter = None
        self.talisman = None

        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app, **kwargs):
        """Initialize application object.

        :param app: An instance of :class:`~flask.Flask`.
        """
        # Init the configuration
        self.init_config(app)

        # Enable secure HTTP headers
        if app.config["APP_ENABLE_SECURE_HEADERS"]:
            self.talisman = Talisman(
                app, **app.config.get("APP_DEFAULT_SECURE_HEADERS", {})
            )

        app.jinja_env.filters["safe_redirect"] = safe_redirect

        # Enable Rate limiter
        # Flask limiter needs to be initialised after talisman, since if the
        # talisman preprocessor doesn't run you will get an error on it's
        # afterprocessor.
        self.limiter = Limiter(
            app,
            key_func=obj_or_import_string(
                app.config.get("RATELIMIT_KEY_FUNC"), default=useragent_and_ip_limit_key
            ),
        )

        # Enable PING view
        if app.config["APP_HEALTH_BLUEPRINT_ENABLED"]:
            blueprint = Blueprint("invenio_app_ping", __name__)
            limiter = self.limiter

            @blueprint.route("/ping", methods=["HEAD", "OPTIONS", "GET"])
            @limiter.exempt
            def ping():
                """Load balancer ping view."""
                return "OK"

            ping.talisman_view_options = {"force_https": False}

            app.register_blueprint(blueprint)

        requestid_header = app.config.get("APP_REQUESTID_HEADER")
        if requestid_header:

            @app.before_request
            def set_request_id():
                """Extracts a request id from an HTTP header."""
                request_id = request.headers.get(requestid_header)
                if request_id:
                    # Capped at 200 to protect against malicious clients
                    # sending very large headers.
                    g.request_id = request_id[:200]

        # If installed register the Flask-DebugToolbar extension
        try:
            from flask_debugtoolbar import DebugToolbarExtension

            app.extensions["flask-debugtoolbar"] = DebugToolbarExtension(app)
        except ImportError:
            app.logger.debug("Flask-DebugToolbar extension not installed.")

        # Force host header check (by evaluating request.host) in order to make
        # Werkzeugs trusted host feature work properly.
        if app.config["APP_ALLOWED_HOSTS"]:

            @app.before_request
            def before_request():
                request.host

        # Add theme template loader
        if app.config.get("APP_THEME"):
            app.jinja_env.loader = ThemeJinjaLoader(app, app.jinja_env.loader)

        # Register self
        app.extensions["invenio-app"] = self

    def init_config(self, app):
        """Initialize configuration.

        :param app: An instance of :class:`~flask.Flask`.
        """
        # RATELIMIT_STORAGE_URL was deprecated in 2.0 and removed in 3.0
        if "RATELIMIT_STORAGE_URL" in app.config:
            app.config["RATELIMIT_STORAGE_URI"] = app.config["RATELIMIT_STORAGE_URL"]
            warnings.warn(
                "RATELIMIT_STORAGE_URL has been renamed to RATELIMIT_STORAGE_URI.",
                DeprecationWarning,
            )

        config_apps = ["APP_", "RATELIMIT_"]
        flask_talisman_debug_mode = "'unsafe-inline'"
        for k in dir(config):
            if any([k.startswith(prefix) for prefix in config_apps]):
                app.config.setdefault(k, getattr(config, k))

        if app.config["DEBUG"]:
            app.config.setdefault("APP_DEFAULT_SECURE_HEADERS", {})
            headers = app.config["APP_DEFAULT_SECURE_HEADERS"]

            # ensure `content_security_policy` is not set to {}
            if headers.get("content_security_policy") != {}:
                headers.setdefault("content_security_policy", {})
                csp = headers["content_security_policy"]

                # ensure `default-src` is not set to []
                if csp.get("default-src") != []:
                    csp.setdefault("default-src", [])

                    # add default `content_security_policy` value when debug
                    if flask_talisman_debug_mode not in csp["default-src"]:
                        csp["default-src"].append(flask_talisman_debug_mode)
